#![allow(dead_code)]
use std::ops::{Add, Sub};

use crate::vector::Vector;
use crate::wide4::f32x4;

/// A position in 3D space.
#[derive(Debug, Copy, Clone)]
#[repr(transparent)]
pub struct Point(pub(crate) f32x4);

impl Point {
    #[inline(always)]
    pub fn new(x: f32, y: f32, z: f32) -> Self {
        Self(f32x4::new(x, y, z, 0.0))
    }

    #[inline(always)]
    pub fn min(self, other: Self) -> Self {
        Self(self.0.min(other.0))
    }

    #[inline(always)]
    pub fn max(self, other: Self) -> Self {
        Self(self.0.max(other.0))
    }

    #[inline(always)]
    pub fn into_vector(self) -> Vector {
        Vector(self.0)
    }

    #[inline(always)]
    pub fn x(self) -> f32 {
        self.0.a()
    }

    #[inline(always)]
    pub fn y(self) -> f32 {
        self.0.b()
    }

    #[inline(always)]
    pub fn z(self) -> f32 {
        self.0.c()
    }

    #[inline(always)]
    #[must_use]
    pub fn set_x(self, x: f32) -> Self {
        Self(self.0.set_a(x))
    }

    #[inline(always)]
    #[must_use]
    pub fn set_y(self, y: f32) -> Self {
        Self(self.0.set_b(y))
    }

    #[inline(always)]
    #[must_use]
    pub fn set_z(self, z: f32) -> Self {
        Self(self.0.set_c(z))
    }
}

impl Add<Vector> for Point {
    type Output = Self;

    #[inline(always)]
    fn add(self, other: Vector) -> Self {
        Self(self.0 + other.0)
    }
}

impl Sub for Point {
    type Output = Vector;

    #[inline(always)]
    fn sub(self, other: Self) -> Vector {
        Vector(self.0 - other.0)
    }
}

impl Sub<Vector> for Point {
    type Output = Self;

    #[inline(always)]
    fn sub(self, other: Vector) -> Self {
        Self(self.0 - other.0)
    }
}

// impl Mul<Transform> for Point {
//     type Output = Self;

//     #[inline]
//     fn mul(self, other: Transform) -> Self {
//         Self {
//             co: other.0.transform_point3a(self.0),
//         }
//     }
// }

//-------------------------------------------------------------

// #[cfg(test)]
// mod tests {
//     use super::super::{Transform, Vector};
//     use super::*;

//     #[test]
//     fn add() {
//         let p1 = Point::new(1.0, 2.0, 3.0);
//         let v1 = Vector::new(1.5, 4.5, 2.5);
//         let p2 = Point::new(2.5, 6.5, 5.5);

//         assert_eq!(p2, p1 + v1);
//     }

//     #[test]
//     fn sub() {
//         let p1 = Point::new(1.0, 2.0, 3.0);
//         let p2 = Point::new(1.5, 4.5, 2.5);
//         let v1 = Vector::new(-0.5, -2.5, 0.5);

//         assert_eq!(v1, p1 - p2);
//     }

//     #[test]
//     fn mul_matrix_1() {
//         let p = Point::new(1.0, 2.5, 4.0);
//         let m = Transform::new_from_values(
//             1.0, 2.0, 2.0, 1.5, 3.0, 6.0, 7.0, 8.0, 9.0, 2.0, 11.0, 12.0,
//         );
//         let pm = Point::new(15.5, 54.0, 70.0);
//         assert_eq!(p * m, pm);
//     }

//     #[test]
//     fn mul_matrix_2() {
//         let p = Point::new(1.0, 2.5, 4.0);
//         let m = Transform::new_from_values(
//             1.0, 2.0, 2.0, 1.5, 3.0, 6.0, 7.0, 8.0, 9.0, 2.0, 11.0, 12.0,
//         );
//         let pm = Point::new(15.5, 54.0, 70.0);
//         assert_eq!(p * m, pm);
//     }

//     #[test]
//     fn mul_matrix_3() {
//         // Make sure matrix multiplication composes the way one would expect
//         let p = Point::new(1.0, 2.5, 4.0);
//         let m1 = Transform::new_from_values(
//             1.0, 2.0, 2.0, 1.5, 3.0, 6.0, 7.0, 8.0, 9.0, 2.0, 11.0, 12.0,
//         );
//         let m2 =
//             Transform::new_from_values(4.0, 1.0, 2.0, 3.5, 3.0, 6.0, 5.0, 2.0, 2.0, 2.0, 4.0, 12.0);
//         println!("{:?}", m1 * m2);

//         let pmm1 = p * (m1 * m2);
//         let pmm2 = (p * m1) * m2;

//         assert!((pmm1 - pmm2).length2() <= 0.00001); // Assert pmm1 and pmm2 are roughly equal
//     }
// }
