#![allow(dead_code)]

use std::ops::{Add, Mul};

use approx::relative_eq;

use crate::point::Point;
use crate::transform_dual::TransformDual;
use crate::wide4::f32x4;

/// An affine transform.
#[derive(Debug, Copy, Clone)]
#[repr(C)]
pub struct Transform {
    pub(crate) m: [f32x4; 3], // Linear matrix.
    pub(crate) t: f32x4,      // Translation.
}

impl Transform {
    /// Creates a new affine transform the specified values:
    ///
    /// ```
    /// a d g j
    /// b e h k
    /// c f i l
    /// ```
    ///
    /// Where j, k, and l are the xyz translation component.
    #[inline]
    #[allow(clippy::many_single_char_names)]
    #[allow(clippy::too_many_arguments)]
    pub fn new(
        a: f32,
        b: f32,
        c: f32,
        d: f32,
        e: f32,
        f: f32,
        g: f32,
        h: f32,
        i: f32,
        j: f32,
        k: f32,
        l: f32,
    ) -> Self {
        Self {
            m: [
                f32x4::new(a, b, c, 0.0),
                f32x4::new(d, e, f, 0.0),
                f32x4::new(g, h, i, 0.0),
            ],
            t: f32x4::new(j, k, l, 0.0),
        }
    }

    /// Creates a new identity transform.
    #[inline]
    pub fn identity() -> Self {
        Self {
            m: [
                f32x4::new(1.0, 0.0, 0.0, 0.0),
                f32x4::new(0.0, 1.0, 0.0, 0.0),
                f32x4::new(0.0, 0.0, 1.0, 0.0),
            ],
            t: f32x4::splat(0.0),
        }
    }

    #[inline]
    pub fn from_location(loc: Point) -> Transform {
        Self {
            m: [
                f32x4::new(1.0, 0.0, 0.0, 0.0),
                f32x4::new(0.0, 1.0, 0.0, 0.0),
                f32x4::new(0.0, 0.0, 1.0, 0.0),
            ],
            t: loc.0,
        }
    }

    /// Returns whether the matrices are approximately equal to each other.
    /// Each corresponding element in the matrices cannot have a relative
    /// error exceeding epsilon.
    #[inline]
    pub fn aprx_eq(&self, other: Transform, epsilon: f32) -> bool {
        let mut eq = true;
        for (t1, t2) in self
            .m
            .iter()
            .chain(&[self.t])
            .zip(other.m.iter().chain(&[other.t]))
        {
            eq &= relative_eq!(t1.a(), t2.a(), epsilon = epsilon);
            eq &= relative_eq!(t1.b(), t2.b(), epsilon = epsilon);
            eq &= relative_eq!(t1.c(), t2.c(), epsilon = epsilon);
        }
        eq
    }

    /// Returns the dual transform, which can do inverse transforms.
    #[inline]
    pub fn compute_dual(self) -> TransformDual {
        TransformDual {
            m: self.m,
            m_inv: f32x4::invert_3x3(self.m),
            t: self.t,
        }
    }

    /// Slower but precise version of `compute_dual()`.
    #[inline]
    pub fn compute_dual_precise(self) -> TransformDual {
        TransformDual {
            m: self.m,
            m_inv: f32x4::invert_3x3_precise(self.m),
            t: self.t,
        }
    }
}

impl Default for Transform {
    fn default() -> Self {
        Self::identity()
    }
}

// /// Multiply two matrices together
// impl Mul for Transform {
//     type Output = Self;

//     #[inline]
//     fn mul(self, rhs: Self) -> Self {
//         Self(rhs.0 * self.0)
//     }
// }

/// Multiply a matrix by a f32
impl Mul<f32> for Transform {
    type Output = Self;

    #[inline]
    fn mul(self, rhs: f32) -> Self {
        Self {
            m: [self.m[0] * rhs, self.m[1] * rhs, self.m[2] * rhs],
            t: self.t * rhs,
        }
    }
}

/// Add two matrices together
impl Add for Transform {
    type Output = Self;

    #[inline]
    fn add(self, rhs: Self) -> Self {
        Self {
            m: [
                self.m[0] + rhs.m[0],
                self.m[1] + rhs.m[1],
                self.m[2] + rhs.m[2],
            ],
            t: self.t + rhs.t,
        }
    }
}

//-------------------------------------------------------------

// #[cfg(test)]
// mod tests {
//     use super::*;

//     #[test]
//     fn equality_test() {
//         let a = Transform::new();
//         let b = Transform::new();
//         let c =
//             Transform::new_from_values(1.1, 0.0, 0.0, 0.0, 0.0, 1.1, 0.0, 0.0, 0.0, 0.0, 1.1, 0.0);

//         assert_eq!(a, b);
//         assert!(a != c);
//     }

//     #[test]
//     fn approximate_equality_test() {
//         let a = Transform::new();
//         let b = Transform::new_from_values(
//             1.000001, 0.0, 0.0, 0.0, 0.0, 1.000001, 0.0, 0.0, 0.0, 0.0, 1.000001, 0.0,
//         );
//         let c = Transform::new_from_values(
//             1.000003, 0.0, 0.0, 0.0, 0.0, 1.000003, 0.0, 0.0, 0.0, 0.0, 1.000003, 0.0,
//         );
//         let d = Transform::new_from_values(
//             -1.000001, 0.0, 0.0, 0.0, 0.0, -1.000001, 0.0, 0.0, 0.0, 0.0, -1.000001, 0.0,
//         );

//         assert!(a.aprx_eq(b, 0.000001));
//         assert!(!a.aprx_eq(c, 0.000001));
//         assert!(!a.aprx_eq(d, 0.000001));
//     }

//     #[test]
//     fn multiply_test() {
//         let a = Transform::new_from_values(
//             1.0, 2.0, 2.0, 1.5, 3.0, 6.0, 7.0, 8.0, 9.0, 2.0, 11.0, 12.0,
//         );
//         let b = Transform::new_from_values(
//             1.0, 5.0, 9.0, 13.0, 2.0, 6.0, 10.0, 14.0, 3.0, 7.0, 11.0, 15.0,
//         );
//         let c = Transform::new_from_values(
//             97.0, 50.0, 136.0, 162.5, 110.0, 60.0, 156.0, 185.0, 123.0, 70.0, 176.0, 207.5,
//         );

//         assert_eq!(a * b, c);
//     }

//     #[test]
//     fn inverse_test() {
//         let a = Transform::new_from_values(
//             1.0, 0.33, 0.0, -2.0, 0.0, 1.0, 0.0, 0.0, 2.1, 0.7, 1.3, 0.0,
//         );
//         let b = a.inverse();
//         let c = Transform::new();

//         assert!((dbg!(a * b)).aprx_eq(dbg!(c), 0.0000001));
//     }
// }
